#version 150
precision highp float; precision highp int;
//Ethan Alexander Shulman 2022 "Chrono Matter" September 10, 2022

in vec3 in_Position;
out vec4 uv;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};


//config
#define ITERATIONS 100
#define EPSILON 0.01
//#define EPSILON_SCALE 1.04
//#define DISTANCE_STEP 1.
#define STEP_DITHER .5
//#define VOLUME_EPSILON 1e-1
#define LIGHTING_EPSILON 1e-4
//#define MAX_BOUNCES 6.
#define LOW_DISCREPANCY 1

#define CULLING_BOUNDS(p) length(p)<120.

//#define LIGHT_IMPORTANCE .25
#define LIGHT_DIRECTION normalize(vec3(1))
#define LIGHT_SIZE .2
#define LIGHT_COLOR vec3(1.75)
#define LIGHT_RESCALE_MAX 1e5
#define LIGHT_CAUSTICS 100.
#define LIGHT_MASK(p) step(4.,p.y)

//#define FOG .6
#define FOG_COLOR(p,d) vec3(.01)
#define FOG_EXP_DISTRIBUTION 1
#define FOG_LIT(p,d) vec3(1)
#define FOG_LIT_ITERATIONS 40
#define FOG_LIT_EPSILON 1e-1
#define FOG_LIT_MASK(p) LIGHT_MASK(p)
#define FOG_LIT_BACKGROUND 1.



//#define ANIMATION_TIME 60.
#define ANIMATION_SAMPLES 10
//#define ANIMATION_RENDER 1
out float time;


//globals
#define PI 3.141592653589793

#ifdef ANIMATION_TIME
#define SFRAME (frame%ANIMATION_SAMPLES)
#else
#define SFRAME frame
#endif


//utility functions
mat2 r2d(float a) {
	float c = cos(a), s = sin(a);
	return mat2(c,s,-s,c);
}

uvec4 hash(uvec4 seed) {
    uvec4 h = (0x6A7F8FAAu^seed)*0x01000193u;
    h = ((h.wxyz>>3u)^h^seed.yzwx)*0x01000193u;
    h = ((h.zwxy>>8u)^h^seed.wxyz)*0x01000193u;
    return h^(h>>11u);
}
#define I2F (1./float(0xFFFFFFFFu))


float line(vec3 p, vec3 a, vec3 b) {
	vec3 pa = p-a, ba = b-a;
	return length(pa-ba*clamp(dot(pa,ba)/dot(ba,ba),0.,1.));
} 


//scene definition

/*returns camera data as mat3 in the following structure {
vec3 position;
float roll, pitch, yaw;
float fieldOfViewTangent, viewNear, orbitDistance;*/
mat3 camera() {
	//animation
	float ang = time/20.*PI*2.;
	return mat3(cos(ang)*10., -20.+time*40./60., sin(ang)*10.-100.+clamp(time/10.,0.,1.)*100.+clamp(time/10.-5.,0.,1.)*50.,
				0,0,0,
				1,1,0);

	//screenshot 0
	//return mat3(0.6,-0.5,-70.,
	//		0,0,0,
	//		1.4,0,0);
	
	//screenshot 1
	//return mat3(-5,3,24,
	//		0,0,PI,
	//		1.,0,0);
}

//returns vec2(distance, materialId) of nearest surface
vec2 geometry(vec3 p) {
	//spheres
	vec3 idp = floor(p/2.+200.)-200.,
		rnd = fract((dot(idp,idp.yzx)+dot(idp,idp.zxy))*vec3(.289768623,.474376,.3388745));
	p = abs(p);
	vec3 rep = mod(p,2.)-1.;
	vec2 geo = vec2(1,0);
	if (rnd.x > pow(length(idp)/20.,.4)) {
		float a1 = rnd.y*20.+time, a2 = rnd.z*20.+time*.18723876,
			s2 = sin(a2);
		geo.x = min(length(rep)-.5, line(rep,vec3(0),vec3(cos(a1)*s2,sin(a1)*s2,cos(a2))*.7)-.3);
	}
	geo.x = max(geo.x,length(p)-42.);

	return geo;
}
vec2 geometryDetail(vec3 p) {
	return geometry(p);
}

/*returns material as mat3 in the following structure {
vec3 color;
float roughness, emission, opacity;*/
mat2x3 material(vec3 p, float mat) {
	float len = length(p)/40.;
	return mat2x3(pow(mix(vec3(1,0,0),mix(vec3(1,1,0),vec3(0,0,1),clamp(len*2.-.8,0.,1.)),clamp(len*4.-.5,0.,1.)),vec3(2.2)), .04,1./(.1+len*len*100.),1);
}


//returns background color at point and direction
vec3 background(vec3 p, vec3 d) {
	return vec3(0);
}
vec3 backgroundLighting(vec3 p, vec3 d) {
	return vec3(.2)*step(42.,length(p));
}


//sample detailed normal using derivatives from geometryDetail
vec3 normal(vec3 p) {
	vec3 n = normalize(vec3(geometryDetail(p+vec3(LIGHTING_EPSILON,0,0)).x,geometryDetail(p+vec3(0,LIGHTING_EPSILON,0)).x,geometryDetail(p+vec3(0,0,LIGHTING_EPSILON)).x)-geometryDetail(p).x);
	if (n != n) return vec3(0,1,0);
	return n;
}


//bidirectional path tracer
uniform sampler2D Buffer0;

void vert() {
	#ifdef ANIMATION_TIME
	#ifdef ANIMATION_RENDER
	time = float(frame)/float(framerate*ANIMATION_SAMPLES);
	#else
	time = mod(float(frame)/float(framerate),ANIMATION_TIME);
	#endif
	#else
	time = 0.;
	#endif
	uv = vec4(in_Position.xy*vec2(screenX/screenY,1),(in_Position.xy*.5+.5)*vec2(screenX,screenY));
	gl_Position = vec4(in_Position,1);
}
void frag() {
	//setup camera ray
	mat3 cam = camera();
	vec4 random = hash(uvec4(SFRAME)*uvec4(1,10,100,1000))*I2F;
	vec3 rp = vec3(0,0,cam[2].z),
		rd = normalize(vec3(uv.xy+(random.xy*2.-1.)/screenY,cam[2].x));
	if (cam[1].x != 0.) {
		mat2 m = r2d(cam[1].x);
		rp.xy *= m; rd.xy *= m;
	}
	if (cam[1].y != 0.) {
		mat2 m = r2d(cam[1].y);
		rp.yz *= m; rd.yz *= m;
	}
	if (cam[1].z != 0.) {
		mat2 m = r2d(cam[1].z);
		rp.xz *= m; rd.xz *= m;
	}
	rp += cam[0].xyz+rd*cam[2].y;
	
	//path trace
	vec3 c = vec3(1);//ray color tint
	vec4 l = vec4(0);//l.xyz = light color accumulation, l.w = number of bounces or -1 if important ray
	float epsilon = EPSILON;
	#ifdef FOG
	vec3 fogPointA = rp, fogPointB;
	#endif
	for (int i = 0; i < ITERATIONS && CULLING_BOUNDS(rp); i++) {
		vec2 geo = geometry(rp);
		if (geo.x <= 0.) {//hit/inside geometry
			mat2x3 mat = material(rp,geo.y);//sample material
			
			#ifdef VOLUME_EPSILON
			if (random.w < mat[1].z) {//stochastic opacity
			#endif
			
				#ifdef LIGHT_IMPORTANCE
				if (l.w < 0.) {	//failed importance cast
					c = vec3(0);
					break;
				}
				#endif
			
				rp -= rd*epsilon*2.;//back up ray
				c *= mat[0];//apply material color
				
				#ifdef LIGHT_IMPORTANCE
				l += vec4(c.xyz*mat[1].y/max(1./LIGHT_RESCALE_MAX,pow(1.-LIGHT_IMPORTANCE,l.w)),1);//emissive lighting + bounce count
				#else
				l += vec4(c.xyz*mat[1].y,1);//emissive lighting + bounce count
				#endif
				
				#ifdef FOG
				if (l.w == 1.) fogPointB = rp;
				#endif
				
				#ifdef MAX_BOUNCES
				if (l.w >= MAX_BOUNCES) break;
				#endif
				
				random = hash(uvec4(uv.zw,l.w,SFRAME))*I2F;
				#ifdef LOW_DISCREPANCY
				if (l.w == 1.) {
					const float r2d = 1.32471795724474602596;
					random.xy = fract(fract((1.+fract(uv.z*.187694)-fract(uv.w*.261867894))*327.78628764)*111.33761235+
									float(SFRAME)*1./vec2(r2d,r2d*r2d));
				}
				#endif
				
				vec3 nrm = normal(rp);
				#ifdef LIGHT_IMPORTANCE
				bool dirLight = random.w < LIGHT_IMPORTANCE;
				vec3 tang = normalize(cross(abs(nrm.y)>(1.-1e-5)?vec3(0,0,-1):vec3(0,-1,0),nrm)), bitang = cross(tang,nrm);
				float cs = pow(random.x,dirLight?mat[1].x*max(0.,.5-mat[1].x):mat[1].x), sn = sqrt(1.-cs*cs), ang = random.y*PI*2.;
				rd = reflect(rd,normalize(mat3(nrm,tang,bitang)*vec3(cs,sn*cos(ang),sn*sin(ang))));//randomly bounce ray off microsurface bump
				
				if (dirLight) {//directional light importance sampling
					float edge = (1.-LIGHT_SIZE*.1)-mat[1].x*4.,
						brdf = clamp((dot(LIGHT_DIRECTION,rd)-edge)/(1.-min(1.,edge)),0.,1.);
					rd = normalize(LIGHT_DIRECTION+(random.xyz-.5)*LIGHT_SIZE);
					if (dot(rd,nrm) < -.2 || brdf < 1e-5) {
						c = vec3(0);
						break;
					}
					c *= brdf/(1.+mat[1].x*LIGHT_CAUSTICS);
					if (l.w > 1.) c *= 1./max(1./LIGHT_RESCALE_MAX,pow(1.-LIGHT_IMPORTANCE,l.w-1.));//rescale brightness of bounces
					l.w = -1.;//mark ray important
					
				}
				#else
				vec3 tang = normalize(cross(abs(nrm.y)>(1.-1e-5)?vec3(0,0,-1):vec3(0,-1,0),nrm)), bitang = cross(tang,nrm);
				float cs = pow(random.x,mat[1].x), sn = sqrt(1.-cs*cs), ang = random.y*PI*2.;
				rd = reflect(rd,normalize(mat3(nrm,tang,bitang)*vec3(cs,sn*cos(ang),sn*sin(ang))));//randomly bounce ray off microsurface bump
				#endif
				
				geo.x = 0.;
			#ifdef VOLUME_EPSILON
			} else {
				geo.x = VOLUME_EPSILON;
			}
			#endif
		} else {
			#ifdef DISTANCE_STEP
			geo.x *= DISTANCE_STEP;
			#endif
		}
			
		//ray march
#ifdef STEP_DITHER
		float stepSz = (epsilon+geo.x)*((1.-STEP_DITHER)+random.w*STEP_DITHER);
#else
		float stepSz = epsilon+geo.x;
#endif
		rp += rd*stepSz;
		random.w = fract(random.w+fract(stepSz)*2763.198735189131);
#ifdef EPSILON_SCALE
		epsilon *= EPSILON_SCALE;
#endif
	}
	
	#ifdef LIGHT_IMPORTANCE
	if (l.w < 0.) {
		l.xyz += c*LIGHT_COLOR*(1.+LIGHT_CAUSTICS)*LIGHT_MASK(rp)/LIGHT_IMPORTANCE;
	} else {
		if (l.w > 0.) c *= 1./max(1./LIGHT_RESCALE_MAX,pow(1.-LIGHT_IMPORTANCE,l.w));//rescale brightness of bounces
	#endif
	//background
	l.xyz += c*(l.w>0.?backgroundLighting(rp,rd):background(rp,rd));
	#ifdef LIGHT_IMPORTANCE
	}
	#endif
	
	#ifdef FOG
	vec3 fpos, fdir;
	#ifdef FOG_LIT
	float flitScale = FOG_LIT_BACKGROUND;
	#endif
	if (l.w == 0.) {
		#ifdef FOG_EXP_DISTRIBUTION
		float flen = length(rp-fogPointA), fdens = min(80.,flen*FOG);
		fpos = mix(fogPointA,rp,log(1./(1.-random.w*(1.-1./exp(fdens))))/fdens);
		#else
		fpos = mix(fogPointA,rp,random.w);
		#endif
		fdir = rd;
	} else {
		fdir = fogPointB-fogPointA;
		float flen = length(fdir), fdens = min(80.,flen*FOG), fv = 1./exp(fdens);
		#ifdef FOG_EXP_DISTRIBUTION
		fpos = fogPointA+fdir*(log(1./(1.-random.w*(1.-fv)))/fdens);
		#else
		fpos = fogPointA+fdir*random.w;
		#endif
		fdir *= 1./max(1e-5,flen);
		l.xyz = mix(FOG_COLOR(fpos,fdir),l.xyz,fv);
		#ifdef FOG_LIT
		flitScale = 1.-fv;
		#endif
	}
	#ifdef FOG_LIT
	vec3 flit = FOG_LIT(fpos,fdir)*flitScale;
	fpos += LIGHT_DIRECTION*FOG_LIT_EPSILON;
	for (int i = 0; i < FOG_LIT_ITERATIONS; i++) {
		vec2 geo = geometry(fpos);
		if (geo.x < FOG_LIT_EPSILON) {
			flit = vec3(0);
			break;
		}
		fpos += LIGHT_DIRECTION*geo.x;
	}
	l.xyz += flit*FOG_LIT_MASK(fpos);
	#endif
	#endif
	
	//sum light in buffer
	vec4 sum = texelFetch(Buffer0,ivec2(uv.zw),0);
	if (SFRAME == 0) sum = vec4(0);
	gl_FragColor = sum+vec4(l.xyz,1);
}
